
/*
 *
 * (C) 2004-2005 Luc Saillard (luc@saillard.org)
 *
 */

#include <linux/videodev.h>
#include <linux/videodev2.h>

#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "libpwc.h"
#include "pwc-dec23.h"

/**
 * libpwc_init - alloc and initialize the private structure used by the library
 *
 * @return NULL error (not enough memory)
 *         pointer to a private structure use between call
 *
 */
struct pwc_t *libpwc_init(void)
{
  struct pwc_t *pwc;

  pwc = calloc(1, sizeof(struct pwc_t));
  if (pwc == NULL)
    return NULL;

  return pwc;
}

/**
 * libpwc_open - open the webcam device
 *
 * @notimplemented
 *
 * @return <0 error
 */
int libpwc_open(struct pwc_t *pwc, const unsigned char *devname)
{
  return -ENOSYS;
}

/**
 * libpwc_close - close an open device
 *
 * @notimplemented
 *
 * @return <0 error
 */
int libpwc_close(struct pwc_t *pwc)
{
  free(pwc);
  return 0;
}

/**
 * libpwc_release - free all memory use the library
 *
 * @return <0 error
 */
int libpwc_release(struct pwc_t *pwc)
{
  pwc_dec23_exit(pwc);
  return 0;
}



/**
 * libpwc_decode - uncompress and convert an image
 *
 * @param pwc: handle on an already initialize pointer
 * @param raw_format: used to determinate the decompressor to use.
 * @param rawstream: the stream to uncompress
 * @param out_format: can be V4L2_PIX_FMT_YUV420 or V4L2_PIX_FMT_RGB24
 * @param buffer: address of an already allocated buffer. If you want the
 * 	library to allocate the buffer set it to NULL. In this case, the buffer
 * 	need to be free using free().
 *
 *  @return <0 error
 *           0 error (buffer need to be free)
 */
int libpwc_decode(struct pwc_t *pwc, 
    		  const int raw_format, const unsigned char *rawstream, 
		  const int out_format, void **buffer)
{
  int yuv_buffer_size;
  int err;

  switch(raw_format)
   {
    case V4L2_PIX_FMT_YUV420:
    case V4L2_PIX_FMT_PWC2:
      break;
    default:
      return -EINVAL;
   }

  yuv_buffer_size = pwc->width * pwc->height * 3 / 2;
  switch (out_format)
   {
    case V4L2_PIX_FMT_RGB24:
    case V4L2_PIX_FMT_YUV420:
      break;

    default:
      return -EINVAL;
   }


  if (*buffer == NULL)
    *buffer = malloc(yuv_buffer_size);
  if (*buffer == NULL)
    return -ENOMEM;

  switch(raw_format)
   {
    case V4L2_PIX_FMT_PWC2:
      err = pwc_dec23_init(pwc, pwc->type, pwc->cmd);
      if (err < 0) 
       {
	 free(*buffer);
	 *buffer = NULL;
	 return err;
       }
      pwc_dec23_decompress(pwc, rawstream, *buffer);
      break;
    case V4L2_PIX_FMT_PWC1:
      /* err = pwc_dec1_init(pwc, pwc->type, pwc->cmd); */
      break;
    case V4L2_PIX_FMT_YUV420:
      memcpy(*buffer, rawstream, yuv_buffer_size);
      break;
   }

  /* Currently the decompressor only works on YUV420P, so we need a second pass
   * to decode it in RGB */
  if (out_format == V4L2_PIX_FMT_RGB24) 
   {
     void *rgb_data, *yuv_data;
     unsigned char *py, *pu, *pv;
     int output_buffer_size = pwc->width * pwc->height * 3;
     int yuvlen;

     rgb_data = malloc(output_buffer_size);
     if (rgb_data == NULL) 
      {
	free(*buffer);
	*buffer = NULL;
	return -ENOMEM;
      }
     yuvlen = pwc->width * pwc->height;
     yuv_data = *buffer;
     py = yuv_data;
     pu = yuv_data + yuvlen;
     pv = yuv_data + yuvlen + yuvlen/4;

     pwc_yuv420p_to_rgb(pwc->width, pwc->height, py, pu, pv, rgb_data);

     *buffer = rgb_data;
     free(yuv_data);
  }

  return 0;
}

static inline unsigned short ple16_to_cpu(const void *arg)
{
  const unsigned char *data = arg;
  return data[0] | (data[1]<<8);
}

/*
 * Simple version to decode a buffer return by video4linux2
 *
 *  @param width: width of the image
 *  @param height: height of the image
 *  @input_buffer: rawstream returns by the V4L2 interface (including the
 *        header)
 *  @output_format: select the output format (V4L2_PIX_FMT_YUV420 
 *        or V4L2_PIX_FMT_RGB24)
 *  @output_buffer: pointer to dynamically array that will be allocated by the
 *        library. This pointer need to be free using free() in case of sucess.
 *
 *  @return  <0 error
 *            0 ok (output_buffer need to be free)
 */
int libpwc_decode_simple(int width, int height,
                         const void *rawstream,
			 int output_format,
			 void **output_buffer)
{
  struct pwc_t *pwc;
  const struct pwc_raw_frame *frame;
  int type;
  int err;

  pwc = libpwc_init();
  if (pwc == NULL)
    return -1;

  /* Fill pwc structure */
  frame = rawstream;
  type = ple16_to_cpu(&frame->type);
  switch (type)
   {
    case 645:
    case 646:
    case 675:
    default:
      return -EINVAL;

    case 680:
    case 690:
      pwc->codec_version = V4L2_PIX_FMT_PWC2;
      break;

    case 720:
    case 730:
    case 740:
    case 750:
      pwc->codec_version = V4L2_PIX_FMT_PWC2;
      break;
   }
  pwc->type = type;
  pwc->vbandlength = ple16_to_cpu(&frame->vbandlength);
  pwc->width = width;
  pwc->height = height;
  memcpy(pwc->cmd, frame->cmd, 4);

  err = libpwc_decode(pwc, V4L2_PIX_FMT_PWC2, frame->rawframe, output_format, output_buffer);
  free(pwc);
  return err;
}


#define SCALEBITS 10
#define ONE_HALF  (1 << (SCALEBITS - 1))
#define FIX(x)    ((int) ((x) * (1<<SCALEBITS) + 0.5))

#define YUV_TO_RGB1_CCIR(cb1, cr1)\
{\
    cb = (cb1) - 128;\
    cr = (cr1) - 128;\
    r_add = FIX(1.40200*255.0/224.0) * cr + ONE_HALF;\
    g_add = - FIX(0.34414*255.0/224.0) * cb - FIX(0.71414*255.0/224.0) * cr + \
            ONE_HALF;\
    b_add = FIX(1.77200*255.0/224.0) * cb + ONE_HALF;\
}

#define CROP(x) \
  (((x)<=0)?0:(((x)>=255)?255:(x)))

#define YUV_TO_RGB2_CCIR(r, g, b, y1)\
{\
    y = ((y1) - 16) * FIX(255.0/219.0);\
    r = CROP((y + r_add) >> SCALEBITS);\
    g = CROP((y + g_add) >> SCALEBITS);\
    b = CROP((y + b_add) >> SCALEBITS);\
}


#define RGB_OUT(d, r, g, b)\
{\
    (d)[0] = r;\
    (d)[1] = g;\
    (d)[2] = b;\
}

#define LIMIT(x)  ( (x) > 0xffff ? 0xff : ( (x) <= 0xff ? 0 : ( (x) >> 8 ) ) )
/*
 * Note: Work only for image multiple of 2
 */
void pwc_yuv420p_to_rgb(unsigned int width, unsigned int height, 
    			const unsigned char *py,
			const unsigned char *pu,
			const unsigned char *pv,
			unsigned char *d)
{
    const unsigned char *y2_ptr;
    unsigned char *d1, *d2;
    int w, y, cb, cr, r_add, g_add, b_add, width2;
    unsigned int r, g, b;

    width2 = (width + 1) >> 1;

    for(;height >= 2; height -= 2) {
        d1 = d;
        d2 = d + (width*3);
        y2_ptr = py + width;

        for(w = width; w >= 2; w -= 2) {
            YUV_TO_RGB1_CCIR(pu[0], pv[0]);
            /* output 4 pixels */
            YUV_TO_RGB2_CCIR(r, g, b, py[0]);
            RGB_OUT(d1, r, g, b);

            YUV_TO_RGB2_CCIR(r, g, b, py[1]);
            RGB_OUT(d1 + 3, r, g, b);

            YUV_TO_RGB2_CCIR(r, g, b, y2_ptr[0]);
            RGB_OUT(d2, r, g, b);

            YUV_TO_RGB2_CCIR(r, g, b, y2_ptr[1]);
            RGB_OUT(d2 + 3, r, g, b);

            d1 += 2 * 3;
            d2 += 2 * 3;

            py += 2;
            y2_ptr += 2;
            pu++;
            pv++;
        }
        d += 2 * (width*3);
        py += width;
    }
}



