

#include "libpwc.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
//#define _BSD_SOURCE
#include <sys/time.h>
#include <time.h>


/*
 * Substract two timeval, and return the delay in ms
 */
#define get_delay(a, b, result)                        \
  do {                                                  \
    unsigned int tv_sec = (a)->tv_sec - (b)->tv_sec;    \
    int tv_usec = (a)->tv_usec - (b)->tv_usec;          \
    if (tv_usec < 0) {                                  \
      tv_sec--;                                         \
      tv_usec += 1000000;                               \
    }                                                   \
    tv_usec += tv_sec * 1000000;                        \
    result = tv_usec/1000;                              \
  } while (0)

enum output_format
{
  FMT_NONE,
  FMT_TGA,
  FMT_RGB,
  FMT_YUV
};

enum input_format
{
  FMT_RAW,
  FMT_TXT,
};

static int decode_image(int width, int height,
                        enum input_format input_fmt,  const char *input_file,
			enum output_format output_fmt, const char *output_file);
static int read_file(const char *filename, void **buffer, size_t *bufferlen);
static int write_file(const char *filename, const void *buffer, size_t bufferlen);
static int write_tga(const char *filename, void *buffer, unsigned int width, unsigned int height);

static int verbose;
static const char short_options [] = "bI:O:i:o:s:v";
static const struct option
long_options [] = {
   	{ "benchmark",  0		 ,	NULL,		'b' },
        { "fmti",       required_argument,      NULL,           'I' },
        { "fmto",       required_argument,      NULL,           'O' },
        { "input",      required_argument,      NULL,           'i' },
        { "output",     required_argument,      NULL,           'o' },
        { "size",       required_argument,      NULL,           's' },
        { "verbose",    0,		      	NULL,           'v' },
        { 0, 0, 0, 0 }
};

static void usage(const char *progname)
{
  fprintf (stderr,
      "Usage: %s [options]\n\n"
      "Options:\n"
      "-b | --benchmark     Print time to decode the buffer\n"
      "-I | --fmti   FORMAT Format of the input file (raw, yuv)\n"
      "-i | --input  FILE   File to convert\n"
      "-O | --fmto   FORMAT Format of the output file (yuv, rgb, tga)\n"
      "-o | --output FILE   File to write the image\n"
      "-s | --size   WxH    WIDTHxHEIGHT (by default 640x480)\n"
      "-v | --verbose       Increase verbosity of the program\n"
      "\n"
      "Version: "LIBPWC_VER_STRING"\n"
      "Copyright: Luc Saillard\n"
      "Licence: LGPL (contains some code from the ffmpeg project)\n"
      "", progname);
}


int main(int argc, char **argv)
{
  unsigned int width=0, height=0;
  char *input_file=NULL, *output_file=NULL;
  enum output_format output_fmt = FMT_TGA;
  enum input_format input_fmt = FMT_RAW;
  int err, do_bench=0;

  while (1) 
   {
     int index;
     int c;

     c = getopt_long (argc, argv, short_options, long_options, &index);

     if (c == -1)
       break;

     switch (c) 
      {
       case 0: /* getopt_long() flag */
	 break;

       case 'b':
	 do_bench = 1;
	 break;
       case 'i':
	 input_file = optarg;
	 break;

       case 'I':
	 if (memcmp(optarg, "raw", 3) == 0)
	    input_fmt = FMT_RAW;
	 else if (memcmp(optarg, "yuv", 3) == 0)
	    input_fmt = FMT_YUV;
	 else 
	  {
	    fprintf(stderr, "Bad input format specified {%s}\n", optarg);
	    usage(argv[0]);
	    exit(EXIT_FAILURE);
	  }
	 break;

       case 'o':
	 output_file = optarg;
	 break;

       case 'O':
	 if (memcmp(optarg, "tga", 3) == 0)
	   output_fmt = FMT_TGA;
	 else if (memcmp(optarg, "rgb", 3) == 0)
	   output_fmt = FMT_RGB;
	 else if (memcmp(optarg, "yuv", 3) == 0)
	    output_fmt = FMT_YUV;
	 else if (memcmp(optarg, "none", 4) == 0)
	    output_fmt = FMT_NONE;
	 else 
	  {
	    fprintf(stderr, "Bad output format specified {%s}\n", optarg);
	    usage(argv[0]);
	    exit(EXIT_FAILURE);
	  }
	 break;

       case 's':
	 if (sscanf(optarg,"%dx%d", &width, &height)!=2) 
	  {
	    fprintf(stderr, "Bad size specified {%s}\n", optarg);
	    usage(argv[0]);
	    exit(EXIT_FAILURE);
	  }
	 break;
       case 'v':
	 verbose++;
	 break;
       default:
	 usage (argv[0]);
	 exit(EXIT_FAILURE);
      }

   }

  if (input_file == NULL) {
     fprintf(stderr, "Please specify an image to convert\n");
     usage (argv[0]);
     exit(EXIT_FAILURE);
  }
  if (output_file == NULL) {
     fprintf(stderr, "Please specify an image to write\n");
     usage (argv[0]);
     exit(EXIT_FAILURE);
  }

  /*
   * We can specify severall image to convert using a printf compatible
   * filename 
   */
  if (strchr(input_file,'%') && strchr(output_file, '%'))
   {
     char tmp_input_file[1024], tmp_output_file[1024];
     struct timeval start, finish;
     unsigned int i, delay;


     gettimeofday(&start, NULL);
     for (i=0; i<100; i++)
      {
	snprintf(tmp_input_file, sizeof(tmp_input_file), input_file, i);
	snprintf(tmp_output_file, sizeof(tmp_output_file), output_file, i);

	err = decode_image(width, height, input_fmt, tmp_input_file, output_fmt, tmp_output_file);
	if (err < 0)
	  break;
      }
     gettimeofday(&finish, NULL);
     if (do_bench)
      {
	get_delay(&finish, &start, delay);
	printf("total decode times takes: %d.%ds\n", delay/1000, delay%1000);
      }
   }
  else
   {
     struct timeval start, finish;
     unsigned int delay;
     gettimeofday(&start, NULL);
     err = decode_image(width, height, input_fmt, input_file, output_fmt, output_file);
     if (do_bench)
      {
	gettimeofday(&finish, NULL);
	get_delay(&finish, &start, delay);
	printf("decode takes: %d.%ds\n", delay/1000, delay%1000);
      }
   }
  
  return err;
}

static inline unsigned short ple16_to_cpu(const void *arg)
{
  const unsigned char *data = arg;
  return data[0] | (data[1]<<8);
}

static int find_size_yuv(const void *input_file, const size_t input_length, int *width, int *height)
{
  static struct pwc_coord pwc_image_sizes[] =
   {
      { 128,  96, 128*96*3/2 },
      { 160, 120, 160*120*3/2 },
      { 176, 144, 176*144*3/2 },
      { 320, 240, 320*240*3/2 },
      { 352, 288, 352*288*3/2 },
      { 640, 480, 640*480*3/2 },
   };
  unsigned int i;
  for (i=0; i<sizeof(pwc_image_sizes)/sizeof(struct pwc_coord); i++)
   {
     if (pwc_image_sizes[i].size == input_length)
      {
	*width = pwc_image_sizes[i].x;
	*height = pwc_image_sizes[i].y;
	return 0;
      }
   }
  return -EINVAL;
}

static int find_size_raw(const void *input_buffer, const size_t input_length, int *width, int *height)
{
  /* perhaps, i should create a function in the library ? */
  static struct pwc_coord pwc_image_sizes[] =
   {
      { 128,  96, 128*96 },
      { 160, 120, 160*120 },
      { 176, 144, 176*144 },
      { 320, 240, 320*240 },
      { 352, 288, 352*288 },
      { 640, 480, 640*480 },
   };
  int rawlen = input_length - sizeof(struct pwc_raw_frame);
  const struct pwc_raw_frame *frame = input_buffer;
  unsigned int vbandlength = ple16_to_cpu(&frame->vbandlength);
  int i;

  rawlen = rawlen / vbandlength*4; /* each vbandlength contains four lines */
  for (i=0; i<sizeof(pwc_image_sizes)/sizeof(struct pwc_coord); i++)
   {
     if (pwc_image_sizes[i].y == rawlen)
      {
	*width = pwc_image_sizes[i].x;
	*height = pwc_image_sizes[i].y;
	return 0;
      }
   }
  return -EINVAL;
}

static int decode_image(int width, int height,
                        enum input_format input_fmt,  const char *input_file,
			enum output_format output_fmt, const char *output_file)
{
  void *input_buffer, *output_buffer;
  unsigned int input_bufferlen, output_bufferlen;
  unsigned int output_bufferformat;
  int err;

  /* Load the file */
  err = read_file(input_file, &input_buffer, &input_bufferlen);
  if (err < 0)
   {
     fprintf(stderr, "Failed to load image %s\n", input_file);
     return err;
   }

  /* Try to guess the size of the image, with and header is not specified */
  if (width == 0 || height == 0)
   {
     if (input_fmt == FMT_RAW)
       find_size_raw(input_buffer, input_bufferlen, &width, &height);
     else if (input_fmt == FMT_YUV)
       find_size_yuv(input_buffer, input_bufferlen, &width, &height);
     if (verbose)
       printf("Detected size: %dx%d\n", width, height);
   }
  if (width == 0 || height == 0)
   {
     fprintf(stderr, "Error size of the image cannot be guessed\n");
     free(input_buffer);
     return err;
   }

  /* Decompress it using the simple API */
  output_buffer = NULL;
  if (output_fmt == FMT_YUV) 
   {
     output_bufferlen = width * height * 3 / 2;
     output_bufferformat = V4L2_PIX_FMT_YUV420;
   } 
  else
   {
     output_bufferlen = width * height * 3;
     output_bufferformat = V4L2_PIX_FMT_RGB24;
   }

  if (input_fmt == FMT_RAW)
   {
     err = libpwc_decode_simple(width, height, input_buffer, output_bufferformat, &output_buffer);
     if (err < 0)
      {
	fprintf(stderr, "Error while decoding the stream \n");
	return err;
      }
   }
  else if (input_fmt == FMT_YUV) 
   {
     /* Decoding using the advanced API (but not real decoding because we want yuv -> rgb) */
     struct pwc_t *pwc = libpwc_init();
     if (pwc == NULL)
       return -1;
     pwc->width = width;
     pwc->height = height;
     err = libpwc_decode(pwc, V4L2_PIX_FMT_YUV420, input_buffer, output_bufferformat, &output_buffer);
     free(pwc);
     if (err < 0)
      {
	fprintf(stderr, "Error while decoding the stream \n");
	return err;
      }
   }

  /* Save it */
  if (output_fmt == FMT_TGA)
    write_tga(output_file, output_buffer, width, height);
  else if (output_fmt == FMT_YUV || output_fmt == FMT_RGB)
    write_file(output_file, output_buffer, output_bufferlen);

  return 0;
}

static int read_file(const char *filename, void **buffer, size_t *bufferlen)
{
  int fd;
  struct stat sts;

  fd = open(filename, O_RDONLY);
  if (fd<0)
    return -1;

  if (fstat(fd, &sts)<0) {
     close(fd);
     return -1;
  }

  *bufferlen = sts.st_size;
  *buffer = malloc(*bufferlen);
  if (*buffer == NULL) {
     close(fd);
     return -1;
  }

  read(fd, *buffer, *bufferlen);
  close(fd);
  return 0;
}

static int write_file(const char *filename, const void *buffer, size_t bufferlen)
{
  int fd, ret, written;

  fd = open(filename, O_CREAT|O_WRONLY|O_TRUNC, S_IREAD|S_IWRITE);
  if (fd<0)
    return -1;

  written = 0;
  do 
   {
     ret = write(fd, buffer+written, bufferlen-written);
     if (ret<0)
       return ret;
     written += ret;
   } while (written < bufferlen);

  close(fd);
  return 0;
}

/*
 * Warning i swap the RGB pixel info BGR
 */
static int write_tga(const char *filename, void *buffer, unsigned int width, unsigned int height)
{
  unsigned char targaheader[18];
  int fd, ret, written;
  int bufferlen = width * height * 3;
  unsigned char *data;

  memset(targaheader,0,sizeof(targaheader));

  targaheader[12] = (unsigned char) (width & 0xFF);
  targaheader[13] = (unsigned char) (width >> 8);
  targaheader[14] = (unsigned char) (height & 0xFF);
  targaheader[15] = (unsigned char) (height >> 8);
  targaheader[17] = 0x20;    /* Top-down, non-interlaced */
  targaheader[2]  = 2;       /* image type = uncompressed RGB */
  targaheader[16] = 24;

 fd = open(filename, O_CREAT|O_WRONLY|O_TRUNC, S_IREAD|S_IWRITE);
  if (fd<0)
    return -1;

  write(fd, targaheader, sizeof(targaheader));

  data = buffer + bufferlen - 3;
  while (data>(unsigned char *)buffer)
   {
     unsigned char c = data[0];
     data[0] = data[2];
     data[2] = c;
     data-=3;
   }

  written = 0;
  do 
   {
     ret = write(fd, buffer+written, bufferlen-written);
     if (ret<0)
       return ret;
     written += ret;
   } while (written < bufferlen);

  close(fd);
  return 0;
}

